from typing import Literal

from .base import ClientBase, object_to_id, parse_data

from ..enums import MuteContext
from ..objects import Filter, FilterKeyword, FilterStatus, ListDeserializer, Status


class FilterBase(ClientBase):
	async def add_filter_keyword(self,
								filter: Filter | int | str,
								keyword: str,
								whole_word: bool = False) -> FilterKeyword:

		filter = object_to_id(filter)
		data = {
			"keyword": keyword,
			"whole_word": whole_word
		}

		return await self.send("POST", f"/api/v2/filters/{filter}/keywords", FilterKeyword, data)


	async def add_status_to_filter(self,
								filter: Filter | str | int,
								status: Status | str | int) -> FilterStatus:

		filter = object_to_id(filter)
		data = {
			"status_id": object_to_id(status)
		}

		return await self.send("POST", f"/api/v2/filters/{filter}/statuses", FilterStatus, data)


	async def delete_filter(self, filter: Filter | int | str) -> None:
		filter = object_to_id(filter)
		await self.send("DELETE", f"/api/v2/filters/{filter}", None)


	async def filter(self, id: int | str) -> Filter:
		return await self.send("GET", f"/api/v2/filters/{id}", Filter)


	async def filters(self) -> list[Filter]:
		return await self.send("GET", "/api/v2/filters", ListDeserializer(Filter))


	async def filter_keyword(self, id: int | str) -> FilterKeyword:
		return await self.send("GET", f"/api/v2/filters/keywords/{id}", FilterKeyword)


	async def filter_keywords(self, id: int | str) -> list[FilterKeyword]:
		return await self.send(
			"GET", "/api/v2/filters/{id}/keywords", ListDeserializer(FilterKeyword)
		)


	async def filtered_status(self, id: int | str) -> FilterStatus:
		return await self.send("GET", f"/api/v2/filters/statuses/{id}", FilterStatus)


	async def filtered_statuses(self, filter: Filter | int | str) -> list[FilterStatus]:
		filter = filter = object_to_id(filter)
		return await self.send(
			"GET", f"/api/v2/filters/{filter}/statuses", ListDeserializer(FilterStatus)
		)


	async def new_filter(self,
						title: str,
						keywords: dict[str, bool],
						context: list[MuteContext | str],
						action: Literal["warn", "hide"] = "warn",
						expires_in: int | None = None) -> Filter:

		# keywords: word, match whole world

		keyword_data = []

		for keyword, whole_word in keywords.items():
			keyword_data.append({
				"keyword": keyword,
				"whole_word": whole_word
			})

		data = parse_data(
			title = title,
			keywords = keywords,
			filter_action = action,
			expires_in = expires_in,
			keywords_attributes = keyword_data
		)

		return await self.send("POST", "/api/v2/filters", Filter, data)


	async def remove_filter_keyword(self, keyword: FilterKeyword | int | str) -> None:
		keyword = object_to_id(keyword)
		await self.send("DELETE", f"/api/v2/filters/keywords/{keyword}", None)


	async def remove_status_from_filter(self, status: FilterStatus | str | int) -> None:
		status = object_to_id(status)
		await self.send("DELETE", f"/api/v2/filters/statuses/{status}", None)


	# todo: handle updating/deleting existing keywords
	async def update_filter(self,
						filter: Filter | int | str,
						title: str | None = None,
						keywords: dict[str, bool] | None = None,
						context: list[MuteContext | str] | None = None,
						action: Literal["warn", "hide"] | None = None,
						expires_in: int | None = None) -> Filter:

		if keywords is not None:
			keyword_data = []

			for keyword, whole_word in keywords.items():
				keyword_data.append({
					"keyword": keyword,
					"whole_word": whole_word
				})

		else:
			keyword_data = None

		filter = object_to_id(filter)
		data = parse_data(
			title = title,
			keywords = keywords,
			filter_action = action,
			expires_in = expires_in,
			keywords_attributes = keyword_data
		)

		return await self.send("PUT", f"/api/v2/filters/{filter}", Filter, data)


	async def update_filter_keyword(self,
									filter_keyword: FilterKeyword | int | str,
									keyword: str,
									whole_word: bool = False) -> FilterKeyword:

		filter_keyword = object_to_id(keyword)
		data = {
			"keyword": keyword,
			"whole_word": whole_word
		}

		return await self.send(
			"PUT", f"/api/v2/filters/keywords/{filter_keyword}", FilterKeyword, data
		)



from blib import Date
from collections.abc import Sequence
from datetime import datetime
from typing import Self


CODE_REDIRECT: str = "urn:ietf:wg:oauth:2.0:oob"


class ApiDate(Date):
	"""
		``Date`` class for API objects
	"""

	FORMAT: str = "%Y-%m-%dT%H:%M:%S.%zZ"
	ALT_FORMATS: Sequence[str] = (
		"%Y-%m-%dT%H:%M:%SZ",
		"%Y-%m-%d"
	)


	@classmethod
	def parse(cls: type[Self], date: datetime | str | int | float, try_iso: bool = True) -> Self:
		if isinstance(date, str):
			date = date.split(".", 1)[0]

		return super().parse(date)



from blib import JsonBase

from .base import ClientBase

from ..objects import Activity, DomainBlock, ExtendedDescription, Instance, ListDeserializer, Rule


class InstanceBase(ClientBase):
	async def domain_blocks(self) -> list[DomainBlock]:
		return await self.send(
			"GET", "/api/v1/instance/domain_blocks", ListDeserializer(DomainBlock)
		)


	async def extended_description(self) -> ExtendedDescription:
		return await self.send(
			"GET", "/api/v1/instance/extended_description", ExtendedDescription, token = False
		)


	async def instance_activity(self) -> list[Activity]:
		return await self.send(
			"GET", "/api/v1/instance/activity", ListDeserializer(Activity), token = False
		)


	async def instance_info(self, version: int = 2) -> Instance:
		# todo: check instance version to get correct endpoint
		return await self.send("GET", f"/api/v{version}/instance", Instance, token = False)


	async def instance_peers(self) -> list[str]:
		return await self.send(
			"GET", "/api/v1/instance/peers", ListDeserializer(None), token = False
		)


	async def rules(self) -> list[Rule]:
		return await self.send(
			"GET", "/api/v1/instance/rules", ListDeserializer(Rule), token = False
		)


	async def translation_languages(self) -> dict[str, list[str]]:
		return await self.send(
			"GET", "/api/v1/instance/translation_languages", JsonBase, token = False
		)



import asyncio
import json

from blib import File, JsonBase, Query, Url, get_object_name
from collections.abc import AsyncIterator, Mapping, Sequence
from httpx import AsyncClient, Response, URL
from typing import Any, IO, Self, TypeVar, Union, overload

from . import api
from .base import CODE_REDIRECT
from .enums import StreamEventType
from .errors import ClientError, ServerError
from .objects import Application, ListDeserializer, ObjectBase


C = TypeVar("C", bound = Union[JsonBase[Any], ObjectBase])


class Client(api.AccountBase, api.AppBase, api.FilterBase, api.InstanceBase,
			api.MiscBase, api.NotificationBase, api.StatusBase, api.StreamBase, api.TagBase,
			api.TimelineBase):
	"Client for the Mastodon API."

	def __init__(self,
				host: str,
				token: str | None = None,
				name: str | None = None,
				website: Url | str | None = None,
				redirect_uri: Url | str | None = None,
				client_id: str | None = None,
				client_secret: str | None = None,
				scopes: Sequence[str] | None = None,
				timeout: int | float = 5,
				stream_timeout: int | float = 600,
				https: bool = True) -> None:
		"""
			Create a new ``Client`` object

			:param host: Hostname to connect to
			:param token: Client token to pass to authenticated endpoints
			:param name: Name to be used in the ``User-Agent`` header and when registering the
				application
			:param website: Website of the application
			:param redirect_uri: Location to redirect to when authenticating a user
			:param client_id: Client identifier from a previous application registration
			:param client_secret: Client secret from a previous application registration
			:param scopes: API scopes the application will use
			:param timeout: Number of seconds to wait for requests before giving up
			:param stream_timeout: Number of seconds to wait for a message or heartbeat on
				streaming endpoints
			:param https: Use ``https`` instead of ``http``
		"""

		self._client = AsyncClient(
			http2 = True,
			timeout = timeout,
			max_redirects = 5,
			headers = {
				"User-Agent": name or "Asynctodon"
			}
		)

		self.https: bool = https
		self.host = host
		self.token: str | None = token
		self.name: str | None = name
		self.website: Url | None = website if website is None else Url.parse(website)
		self.redirect_uri: str = redirect_uri or CODE_REDIRECT
		self.client_id: str | None = client_id
		self.client_secret: str | None = client_secret
		self.scopes: list[str] = ["read"] if scopes is None else list(scopes)
		self.stream_timeout: int | float = stream_timeout


	def __repr__(self) -> str:
		return f"Client({repr(self.host)}, name={repr(self.name)})"


	async def __aenter__(self) -> Self:
		return self


	async def __aexit__(self, *_: Any) -> None:
		await self.close()


	@classmethod
	def new_from_app(cls: type[Self],
					app: Application,
					host: str,
					token: str | None = None) -> Self:
		"""
			Create a new ``Client`` from an ``Application``

			:param app: Application info to use
			:param host: Instance to connect to
			:param token: Access token to use
		"""

		client = cls(host, token)
		client.set_details_from_app(app)
		return client


	@property
	def host(self) -> str:
		"Instance the client will send requests to"

		return self._client._base_url.host


	@host.setter
	def host(self, value: str) -> None:
		if "/" in value:
			raise ValueError("Host must not contain path segments")

		self._client._base_url = URL(f"{'https' if self.https else 'http'}://{value}/")


	async def close(self) -> None:
		"Close all open connections"

		if self._client.is_closed:
			return

		await self._client.aclose()
		await asyncio.sleep(0)


	def set_details_from_app(self, app: Application) -> None:
		"""
			Set application details from an ``Application`` object

			:param app: Application to set details from
		"""

		self.name = app.name
		self.website = app.website
		self.redirect_url = app.redirect_uri or CODE_REDIRECT
		self.client_id = app.client_id
		self.client_secret = app.client_secret
		self.scopes = app.scope or ["read"]


	@overload
	async def send(self,
				method: str,
				path: str,
				cls: type[C],
				data: Mapping[str, Any] | Sequence[Any] | None = None,
				query: Query | Mapping[str, str] | None = None,
				form: Mapping[str, Any] | None = None,
				token: bool = True) -> C:
		...


	@overload
	async def send(self,
				method: str,
				path: str,
				cls: ListDeserializer[C],
				data: Mapping[str, Any] | Sequence[Any] | None = None,
				query: Query | Mapping[str, str] | None = None,
				form: Mapping[str, Any] | None = None,
				token: bool = True) -> list[C]:
		...


	@overload
	async def send(self,
				method: str,
				path: str,
				cls: None,
				data: Mapping[str, Any] | Sequence[Any] | None = None,
				query: Query | Mapping[str, str] | None = None,
				form: Mapping[str, Any] | None = None,
				token: bool = True) -> None:
		...


	async def send(self,
				method: str,
				path: str,
				cls: type[C] | ListDeserializer[C] | None,
				data: Mapping[str, Any] | Sequence[Any] | None = None,
				query: Query | Mapping[str, str] | None = None,
				form: Mapping[str, Any] | None = None,
				token: bool = True) -> C | list[C] | None:
		"""
			Send a request to the instance

			:param method: HTTP method of the endpoint
			:param path: Path of the endpoint
			:param cls: Class to use when parsing returned data. Use ``None`` if not expecting a
				response.
			:param data: JSON data to send with the request. Do not use with ``form``.
			:param query: Query parameters to append to the url
			:param form: HTML form data to send with the request. Do not use with ``data``.
			:param token: Ensure a token is set before sending the request
		"""

		open_files: list[IO[bytes]] = []
		headers: dict[str, str] = {
			"Accept": "application/json"
		}

		if self.token is None:
			if token:
				raise ClientError(f"Token required for endpoint: {method} {path}")

		else:
			headers["Authorization"] = f"Bearer {self.token}"

		body: str | bytes | dict[str, Any] = {}
		files: dict[str, tuple[str, IO[bytes]]] = {}
		resp: Response | None = None

		try:
			if form is not None:
				for key, value in form.items():
					if isinstance(value, File):
						value = value.resolve()
						open_files.append(open(value, "rb"))
						files[key] = (value.name, open_files[-1])

					else:
						body[key] = value # type: ignore[index]

			elif data is not None:
				headers["Content-Type"] = "application/json"
				body = data.to_json() if isinstance(data, JsonBase) else json.dumps(data)

			resp = await self._client.request(
				method = method,
				url = path,
				params = query, # type: ignore[arg-type]
				files = files or None,
				content = body or None, # type: ignore[arg-type]
				headers = headers
			)

			resp_data = (await resp.aread()).decode(resp.charset_encoding or "utf-8")

			if resp.status_code >= 400:
				if resp.headers.get("Content-Type") == "application/json":
					msg: JsonBase[str] = JsonBase.parse(resp_data)
					raise ServerError(resp.status_code, msg.get("error_description", msg["error"]))

				# this should never get called, but keeping here just in case
				if resp.headers.get("Content-Type") == "text/html":
					raise ServerError(resp.status_code, resp.reason_phrase)

				raise ServerError(resp.status_code, resp_data)

			if cls is not None:
				if type(cls) is ListDeserializer:
					return cls(json.loads(resp_data))

				if type(cls) is type and issubclass(cls, JsonBase):
					return cls.parse(resp_data) # type: ignore[return-value]

				raise ValueError(f"Invalid parser class: {get_object_name(cls)}")

		finally:
			if resp is not None:
				await resp.aclose()

			for file in open_files:
				file.close()

		return None


	async def stream(self,
				path: str,
				query: Query | Mapping[str, str] | None = None) -> AsyncIterator[api.StreamEvent]:
		"""
			Connect to a streaming endpoint and return an iterator of events

			:param path: Path to the streaming endpoint
			:param query: Query parameters to append to the url
		"""

		if self.token is None:
			raise ClientError("Missing client token")

		headers: dict[str, str] = {
			"Accept": "application/json",
			"Authorization": f"Bearer {self.token}"
		}

		if query is None:
			query = {}

		pre_resp = self._client.stream(
			"GET", path,
			params = query, # type: ignore[arg-type]
			headers = headers,
			timeout = self.stream_timeout
		)

		event_name: str | None = None

		async with pre_resp as resp:
			async for line in resp.aiter_lines():
				if line.startswith("{\"error\""):
					data: JsonBase[str] = JsonBase.parse(line)
					raise ServerError(400, data["error"])

				if line.startswith("event:"):
					event_name = line.split(":", 1)[1].strip()

				elif line.startswith("data:") and event_name is not None:
					event_type = StreamEventType.parse(event_name)
					line_data = line.split(":", 1)[1].strip()

					if event_name in {"delete", "announcement.delete"}:
						yield api.StreamEvent(event_type, JsonBase({"id": line_data}))

					elif event_name == "filters_changed":
						# not sure what the data for this event looks like, so ignore for now
						yield api.StreamEvent(event_type, JsonBase({"id": "0"}))

					else:
						yield api.StreamEvent(event_type, JsonBase.parse(line_data))

					event_name = None



from blib import JsonBase, Query, Url
from collections.abc import AsyncIterator, Mapping, Sequence
from dataclasses import dataclass
from typing import Any, TypeVar, Union, overload

from ..enums import StreamEventType
from ..objects import Application, ListDeserializer, ObjectBase


T = TypeVar("T")
C = TypeVar("C", bound = Union[JsonBase[Any], ObjectBase])
O = TypeVar("O", bound = ObjectBase) # noqa: E741


@overload
def object_to_id(obj: None, prop: str | None = None) -> None:
	...


@overload
def object_to_id(obj: O | int | str, prop: str | None = None) -> str:
	...


def object_to_id(obj: O | int | str | None, prop: str | None = None) -> str | None:
	"""
		Attempts to return the specified property (`id` by default) of an object. If the object is a
		string, just return it
	"""

	if obj is None:
		return None

	if isinstance(obj, ObjectBase):
		return str(getattr(obj, prop or "id"))

	return str(obj)


def object_to_ids(obj_list: list[O | int | str], prop: str | None = None) -> list[str]:
	"Same as `object_to_id` except for a list of items"

	return [object_to_id(item, prop) for item in obj_list]


def parse_data(**kwargs: Any) -> dict[str, Any] | None:
	"Go though some key/value pairs and remove any pairs that have a value of `None`"

	return {key: value for key, value in kwargs.items() if value is not None} or None


def version(introduced: str, deprecated: str | None, removed: str | None) -> None:
	pass


@dataclass(slots = True)
class StreamEvent:
	"Represents an event from a streaming API endpoint. `data` will always have an `id` key."

	type: StreamEventType
	"Type of event"

	data: JsonBase[Any]
	"Data associated with the event"


class ClientBase:
	host: str
	https: bool
	token: str | None
	name: str | None
	website: Url | None
	redirect_uri: str
	client_id: str | None
	client_secret: str | None
	scopes: list[str]


	def set_details_from_app(self, app: Application) -> None:
		...


	@overload
	async def send(self,
				method: str,
				path: str,
				cls: type[C],
				data: Mapping[str, Any] | Sequence[Any] | None = None,
				query: Query | Mapping[str, str] | None = None,
				form: Mapping[str, Any] | None = None,
				token: bool = True) -> C:
		...


	@overload
	async def send(self,
				method: str,
				path: str,
				cls: ListDeserializer[C],
				data: Mapping[str, Any] | Sequence[Any] | None = None,
				query: Query | Mapping[str, str] | None = None,
				form: Mapping[str, Any] | None = None,
				token: bool = True) -> list[C]:
		...


	@overload
	async def send(self,
				method: str,
				path: str,
				cls: None,
				data: Mapping[str, Any] | Sequence[Any] | None = None,
				query: Query | Mapping[str, str] | None = None,
				form: Mapping[str, Any] | None = None,
				token: bool = True) -> None:
		...


	async def send(self,
				method: str,
				path: str,
				cls: type[C] | ListDeserializer[C] | None,
				data: Mapping[str, Any] | Sequence[Any] | None = None,
				query: Query | Mapping[str, str] | None = None,
				form: Mapping[str, Any] | None = None,
				token: bool = True) -> C | list[C] | None:

		raise NotImplementedError


	def stream(self,
				path: str,
				query: Mapping[str, str] | None = None) -> AsyncIterator[StreamEvent]:

		raise NotImplementedError



from .base import ClientBase, object_to_id, parse_data

from ..objects import List, ListDeserializer, Status, Tag


class TimelineBase(ClientBase):
	async def hashtag_timeline(self,
							tag: Tag | str,
							any_tags: list[str] | None = None,
							all_tags: list[str] | None = None,
							no_tags: list[str] | None = None,
							only_local: bool = False,
							only_remote: bool = False,
							only_media: bool = False,
							min_id: int | None = None,
							max_id: int | None = None,
							since_id: int | None = None,
							limit: int = 20) -> list[Status]:

		tag = object_to_id(tag, "name")
		data = parse_data(
			any = any_tags,
			all = all_tags,
			none = no_tags,
			local = only_local,
			remote = only_remote,
			only_media = only_media,
			min_id = min_id,
			max_id = max_id,
			since_id = since_id,
			limit = limit
		)

		return await self.send(
			"GET", f"/api/v1/timelines/tag/{tag}", ListDeserializer(Status), data
		)


	async def home_timeline(self,
							min_id: int | None = None,
							max_id: int | None = None,
							since_id: int | None = None,
							limit: int = 20) -> list[Status]:

		data = parse_data(
			min_id = min_id,
			max_id = max_id,
			since_id = since_id,
			limit = limit
		)

		return await self.send("GET", "/api/v1/timelines/home", ListDeserializer(Status), data)


	async def list_timeline(self,
							list_id: List | int | str,
							min_id: int | None = None,
							max_id: int | None = None,
							since_id: int | None = None,
							limit: int = 20) -> list[Status]:

		list_id = object_to_id(list_id)
		data = parse_data(
			min_id = min_id,
			max_id = max_id,
			since_id = since_id,
			limit = limit
		)

		return await self.send(
			"GET", f"/api/v1/timelines/list/{list_id}", ListDeserializer(Status), data
		)


	async def public_timeline(self,
							only_local: bool = False,
							only_remote: bool = False,
							only_media: bool = False,
							min_id: int | None = None,
							max_id: int | None = None,
							since_id: int | None = None,
							limit: int = 20) -> list[Status]:

		if only_local and only_remote:
			raise ValueError("only_local and only_remote are mutually exclusive")

		data = parse_data(
			local = only_local,
			remote = only_remote,
			only_media = only_media,
			min_id = min_id,
			max_id = max_id,
			since_id = since_id,
			limit = limit
		)

		return await self.send("GET", "/api/v1/timelines/public", ListDeserializer(Status), data)



import asyncio

from collections.abc import AsyncIterator, Callable
from typing import ParamSpec

from .base import ClientBase, StreamEvent

from ..errors import ServerError


P = ParamSpec("P")


class StreamBase(ClientBase):
	async def stream_health(self) -> bool:
		try:
			await self.send("GET", "/api/v1/streaming/health", None)
			return True

		except ServerError as error:
			if error.status >= 500:
				return False

			raise


	def stream_hashtag(self, tag: str, only_local: bool = False) -> AsyncIterator[StreamEvent]:
		params = {"tag": tag}
		path = "/api/v1/streaming/hashtag"

		if only_local:
			path += "/local"

		return self.stream(path, params)


	def stream_home(self) -> AsyncIterator[StreamEvent]:
		return self.stream("/api/v1/streaming/user")


	def stream_direct(self) -> AsyncIterator[StreamEvent]:
		return self.stream("/api/v1/streaming/direct")


	def stream_list(self, list_id: str) -> AsyncIterator[StreamEvent]:
		params = {"list": list_id}
		return self.stream("/api/v1/streaming/list", params)


	def stream_notifications(self) -> AsyncIterator[StreamEvent]:
		return self.stream("/api/v1/streaming/user/notification")


	def stream_public(self,
					only_media: bool = False,
					only_remote: bool = False,
					only_local: bool = False) -> AsyncIterator[StreamEvent]:

		params = {"only_media": str(only_media).lower()}
		path = "/api/v1/streaming/public"

		if only_remote and only_local:
			raise ValueError("'only_remote' and 'only_local' are mutually exclusive")

		if only_remote:
			path += "/remote"

		elif only_local:
			path += "/local"

		return self.stream(path, params)


	async def stream_with_callback(self,
							iterator: AsyncIterator[StreamEvent],
							callback: Callable[[StreamEvent], None]) -> asyncio.Task[None]:
		"""
			Run a stream iterator in the background and return the associated :class:`asyncio.Task`
			object

			:param iterator: Async iterator from a :meth:`Client.stream` method
			:param callback: Function that gets called for every stream event
		"""

		return asyncio.create_task(self.handle_stream_with_callback(iterator, callback))


	async def handle_stream_with_callback(self,
							iterator: AsyncIterator[StreamEvent],
							callback: Callable[[StreamEvent], None]) -> None:

		try:
			async for event in iterator:
				callback(event)

		except asyncio.CancelledError:
			pass



from .base import StreamEvent

from .accounts import AccountBase
from .apps import AppBase
from .filters import FilterBase
from .instances import InstanceBase
from .misc import MiscBase
from .notifications import NotificationBase
from .statuses import StatusBase
from .streams import StreamBase
from .tags import TagBase
from .timelines import TimelineBase


__all__ = (
	"StreamEvent",

	"AccountBase",
	"AppBase",
	"FilterBase",
	"InstanceBase",
	"MiscBase",
	"NotificationBase",
	"StatusBase",
	"StreamBase",
	"TagBase",
	"TimelineBase"
)
